@@ -31,7 +31,10 @@
 #include <sys/wait.h>
 #endif
 
+#include <unistd.h>
 #include <glib.h>
+#include <glib/gstdio.h>
+#include <fcntl.h>
 #include <gtk/gtk.h>
 
 #include <libxfce4ui/libxfce4ui.h>
@@ -50,11 +53,23 @@
 /* Increase this number if new gtk settings have been added */
 #define INITIALIZE_UINT (1)
 
+#define COLOR_SCHEME_SYMBOL ((gpointer) 1)
+#define GTK_RC_TOKEN_FG ((gpointer) 2)
+#define GTK_RC_TOKEN_BG ((gpointer) 3)
+#define GTK_RC_TOKEN_NORMAL ((gpointer) 4)
+#define GTK_RC_TOKEN_SELECTED ((gpointer) 5)
+
+gchar *gtkrc_get_color_scheme_for_theme (const gchar *gtkrc_filename);
+gboolean color_scheme_parse_colors (const gchar *scheme, GdkColor *colors);
+static GdkPixbuf *theme_create_preview (GdkColor *colors);
+
 enum
 {
+    COLUMN_THEME_PREVIEW,
     COLUMN_THEME_NAME,
     COLUMN_THEME_DISPLAY_NAME,
     COLUMN_THEME_COMMENT,
+    COLUMN_THEME_NO_CACHE,
     N_THEME_COLUMNS
 };
 
@@ -65,6 +80,13 @@ enum
     N_RGBA_COLUMNS
 };
 
+enum {
+	COLOR_FG,
+	COLOR_BG,
+	COLOR_SELECTED_BG,
+	NUM_SYMBOLIC_COLORS
+};
+
 /* String arrays with the settings in combo boxes */
 static const gchar* toolbar_styles_array[] =
 {
@@ -99,6 +121,46 @@ static GOptionEntry option_entries[] =
 /* Global xfconf channel */
 static XfconfChannel *xsettings_channel;
 
+typedef struct
+{
+    GtkListStore *list_store;
+    GtkTreeView *tree_view;
+} preview_data;
+
+
+static preview_data *
+preview_data_new (GtkListStore *list_store,
+                  GtkTreeView *tree_view)
+{
+    preview_data *pd;
+
+    g_return_val_if_fail (list_store != NULL, NULL);
+    g_return_val_if_fail (tree_view != NULL, NULL);
+    g_return_val_if_fail (GTK_IS_LIST_STORE (list_store), NULL);
+    g_return_val_if_fail (GTK_IS_TREE_VIEW (tree_view), NULL);
+
+    pd = g_slice_new0 (preview_data);
+    g_return_val_if_fail (pd != NULL, NULL);
+
+    pd->list_store = list_store;
+    pd->tree_view = tree_view;
+
+    g_object_ref (G_OBJECT (pd->list_store));
+    g_object_ref (G_OBJECT (pd->tree_view));
+
+    return pd;
+}
+
+static void
+preview_data_free (preview_data *pd)
+{
+    if (G_UNLIKELY (pd == NULL))
+        return;
+    g_object_unref (G_OBJECT (pd->list_store));
+    g_object_unref (G_OBJECT (pd->tree_view));
+    g_slice_free (preview_data, pd);
+}
+
 static int
 compute_xsettings_dpi (GtkWidget *widget)
 {
@@ -123,6 +185,307 @@ compute_xsettings_dpi (GtkWidget *widget)
     return dpi;
 }
 
+/* Try to retrieve the color scheme and alternatively parse the colors from the gtkrc */
+gchar *
+gtkrc_get_color_scheme_for_theme (const gchar *gtkrc_filename)
+{
+    gint file = -1;
+    GString *result_string = g_string_new("");
+    GString *fallback_string = g_string_new("");
+    GString *tmp_string = g_string_new("");
+    gchar *result = NULL;
+    gboolean bg_normal = FALSE;
+    gboolean bg_selected = FALSE;
+    gboolean fg_normal = FALSE;
+    GSList *files = NULL;
+    GSList *read_files = NULL;
+    GTokenType token;
+    GScanner *scanner = gtk_rc_scanner_new ();
+
+    g_return_val_if_fail (gtkrc_filename != NULL, "");
+
+    g_scanner_scope_add_symbol (scanner, 0, "gtk_color_scheme", COLOR_SCHEME_SYMBOL);
+    g_scanner_scope_add_symbol (scanner, 0, "gtk-color-scheme", COLOR_SCHEME_SYMBOL);
+    g_scanner_scope_add_symbol (scanner, 0, "fg", GTK_RC_TOKEN_FG);
+    g_scanner_scope_add_symbol (scanner, 0, "bg", GTK_RC_TOKEN_BG);
+    g_scanner_scope_add_symbol (scanner, 0, "NORMAL", GTK_RC_TOKEN_NORMAL);
+    g_scanner_scope_add_symbol (scanner, 0, "SELECTED", GTK_RC_TOKEN_SELECTED);
+
+    files = g_slist_prepend (files, g_strdup (gtkrc_filename));
+    while (files != NULL)
+    {
+        gchar *filename = files->data;
+        files = g_slist_delete_link (files, files);
+
+        if (filename == NULL)
+            continue;
+
+        if (g_slist_find_custom (read_files, filename, (GCompareFunc) strcmp))
+        {
+            g_warning ("Recursion in the gtkrc detected!");
+            g_free (filename);
+            continue; /* skip this file since we've done it before... */
+        }
+
+        read_files = g_slist_prepend (read_files, filename);
+
+        file = g_open (filename, O_RDONLY);
+        if (file == -1)
+            g_warning ("Could not open file \"%s\"", filename);
+        else
+        {
+            g_scanner_input_file (scanner, file);
+            while ((token = g_scanner_get_next_token (scanner)) != G_TOKEN_EOF)
+            {
+                /* Scan the gtkrc file for the gtk-color-scheme */
+                if (GINT_TO_POINTER (token) == COLOR_SCHEME_SYMBOL)
+                {
+                    if (g_scanner_get_next_token (scanner) == '=')
+                    {
+                        token = g_scanner_get_next_token (scanner);
+                        if (token == G_TOKEN_STRING)
+                        {
+                            g_string_append_printf(result_string, "\n%s", scanner->value.v_string);
+                        }
+                        bg_normal = g_strstr_len(result_string->str, -1, "bg_color") != NULL;
+                        bg_selected = g_strstr_len(result_string->str, -1, "selected_bg_color") != NULL;
+                        fg_normal = g_strstr_len(result_string->str, -1, "fg_color") != NULL;
+                    }
+                }
+                /* Scan the gtkrc file for first occurences of bg[NORMAL], bg[SELECTED] and fg[NORMAL] in case
+                 * it doesn't provide a gtk-color-scheme */
+                if (result_string->len == 0)
+                {
+                    if (GINT_TO_POINTER (token) == GTK_RC_TOKEN_BG)
+                    {
+                        if (g_scanner_get_next_token (scanner) == G_TOKEN_LEFT_BRACE)
+                        {
+                            /* bg[SELECTED] */
+                            if (GINT_TO_POINTER (token = g_scanner_get_next_token (scanner)) == GTK_RC_TOKEN_SELECTED
+                                && g_scanner_get_next_token (scanner) == G_TOKEN_RIGHT_BRACE
+                                && g_scanner_get_next_token (scanner) == '=')
+                            {
+                                token = g_scanner_get_next_token (scanner);
+                                /* Parse colors in hex #rrggbb format */
+                                if (token == G_TOKEN_STRING)
+                                {
+                                    if (!g_strrstr (fallback_string->str, "selected_bg_color"))
+                                    {
+                                        g_string_append_printf(fallback_string, "\nselected_bg_color:%s", scanner->value.v_string);
+                                        bg_selected = TRUE;
+                                    }
+                                }
+                                /* Parse colors in { r, g, b } format */
+                                else if (token == G_TOKEN_LEFT_CURLY)
+                                {
+                                    if (!g_strrstr (fallback_string->str, "selected_bg_color"))
+                                    {
+                                        g_string_erase(tmp_string, 0, -1);
+                                        while (token != G_TOKEN_RIGHT_CURLY)
+                                        {
+                                            token = g_scanner_get_next_token (scanner);
+                                            if (token == G_TOKEN_FLOAT)
+                                                g_string_append_printf(tmp_string, "%02x", (uint) (scanner->value.v_float*255));
+                                        }
+                                        g_string_append_printf(fallback_string, "\nselected_bg_color:#%s", tmp_string->str);
+                                        bg_selected = TRUE;
+                                    }
+                                }
+                            }
+                            /* bg[NORMAL] */
+                            else if (GINT_TO_POINTER (token) == GTK_RC_TOKEN_NORMAL
+                                     && g_scanner_get_next_token (scanner) == G_TOKEN_RIGHT_BRACE
+                                     && g_scanner_get_next_token (scanner) == '=')
+                            {
+                                token = g_scanner_get_next_token (scanner);
+                                /* Parse colors in hex #rrggbb format */
+                                if (token == G_TOKEN_STRING)
+                                {
+                                    if (!g_strrstr (fallback_string->str, "bg_color"))
+                                    {
+                                        g_string_append_printf(fallback_string, "\nbg_color:%s", scanner->value.v_string);
+                                        bg_normal = TRUE;
+                                    }
+                                }
+                                /* Parse colors in { r, g, b } format */
+                                else if (token == G_TOKEN_LEFT_CURLY)
+                                {
+                                    if (!g_strrstr (fallback_string->str, "bg_color"))
+                                    {
+                                        g_string_erase(tmp_string, 0, -1);
+                                        while (token != G_TOKEN_RIGHT_CURLY)
+                                        {
+                                            token = g_scanner_get_next_token (scanner);
+                                            if (token == G_TOKEN_FLOAT)
+                                                g_string_append_printf(tmp_string, "%02x", (uint) (scanner->value.v_float*255));
+                                        }
+                                        g_string_append_printf(fallback_string, "\nbg_color:#%s", tmp_string->str);
+                                        bg_normal = TRUE;
+                                    }
+                                }
+                            }
+                        }
+                    }
+                    else if (GINT_TO_POINTER (token) == GTK_RC_TOKEN_FG)
+                    {
+                        /* fg[NORMAL] */
+                        if (g_scanner_get_next_token (scanner) == G_TOKEN_LEFT_BRACE
+                            && GINT_TO_POINTER (token = g_scanner_get_next_token (scanner)) == GTK_RC_TOKEN_NORMAL
+                                && g_scanner_get_next_token (scanner) == G_TOKEN_RIGHT_BRACE
+                                && g_scanner_get_next_token (scanner) == '=')
+                        {
+                            token = g_scanner_get_next_token (scanner);
+                            /* Parse colors in hex #rrggbb format */
+                            if (token == G_TOKEN_STRING)
+                            {
+                                if (!g_strrstr (fallback_string->str, "fg_color"))
+                                {
+                                    g_string_append_printf(fallback_string, "\nfg_color:%s", scanner->value.v_string);
+                                    fg_normal = TRUE;
+                                }
+                            }
+                            /* Parse colors in { r, g, b } format */
+                            else if (token == G_TOKEN_LEFT_CURLY)
+                            {
+                                if (!g_strrstr (fallback_string->str, "fg_color"))
+                                {
+                                    g_string_erase(tmp_string, 0, -1);
+                                    while (token != G_TOKEN_RIGHT_CURLY)
+                                    {
+                                        token = g_scanner_get_next_token (scanner);
+                                        if (token == G_TOKEN_FLOAT)
+                                            g_string_append_printf(tmp_string, "%02x", (uint) (scanner->value.v_float*255));
+                                    }
+                                    g_string_append_printf(fallback_string, "\nfg_color:#%s", tmp_string->str);
+                                    fg_normal = TRUE;
+                                }
+                            }
+                        }
+                    }
+                }
+                /* Check whether we can stop parsing because all colors have been retrieved somehow */
+                if (bg_normal && bg_selected && fg_normal)
+                    break;
+            }
+        close (file);
+        }
+    }
+
+    g_slist_foreach (read_files, (GFunc) g_free, NULL);
+    g_slist_free (read_files);
+
+    g_scanner_destroy (scanner);
+
+    /* Use the fallback colors parsed from the theme if gtk-color-scheme is not defined */
+    /*
+    if (!result)
+        result = fallback;
+    */
+    if (result_string->len == 0)
+        result = g_strdup(fallback_string->str);
+    else
+        result = g_strdup(result_string->str);
+
+    g_string_free(result_string, TRUE);
+    g_string_free(fallback_string, TRUE);
+    g_string_free(tmp_string, TRUE);
+
+    return result;
+}
+
+gboolean
+color_scheme_parse_colors (const gchar *scheme, GdkColor *colors)
+{
+    gchar **color_scheme_strings, **color_scheme_pair, *current_string;
+    gboolean found = FALSE;
+    gint i;
+
+    if (!scheme || !strcmp (scheme, ""))
+        return FALSE;
+
+    /* Initialise the array, fallback color is white */
+    for (i = 0; i < NUM_SYMBOLIC_COLORS; i++)
+        colors[i].red = colors[i].green = colors[i].blue = 65535.0;
+
+    /* The color scheme string consists of name:color pairs, separated by
+     * either semicolons or newlines, so first we split the string up by delimiter */
+    if (g_strrstr (scheme, ";"))
+        color_scheme_strings = g_strsplit (scheme, ";", 0);
+    else
+        color_scheme_strings = g_strsplit (scheme, "\n", 0);
+
+    /* Loop through the name:color pairs, and save the color if we recognise the name */
+    i = 0;
+    while ((current_string = color_scheme_strings[i++]))
+    {
+        color_scheme_pair = g_strsplit (current_string, ":", 0);
+
+        if (color_scheme_pair[0] != NULL && color_scheme_pair[1] != NULL)
+        {
+            g_strstrip (color_scheme_pair[0]);
+            g_strstrip (color_scheme_pair[1]);
+
+            if (!strcmp ("fg_color", color_scheme_pair[0]))
+            {
+              gdk_color_parse (color_scheme_pair[1], &colors[COLOR_FG]);
+              found = TRUE;
+            }
+            else if (!strcmp ("bg_color", color_scheme_pair[0]))
+            {
+              gdk_color_parse (color_scheme_pair[1], &colors[COLOR_BG]);
+              found = TRUE;
+            }
+            else if (!strcmp ("selected_bg_color", color_scheme_pair[0]))
+            {
+              gdk_color_parse (color_scheme_pair[1], &colors[COLOR_SELECTED_BG]);
+              found = TRUE;
+            }
+        }
+
+        g_strfreev (color_scheme_pair);
+    }
+
+    g_strfreev (color_scheme_strings);
+
+    return found;
+}
+
+static GdkPixbuf *
+theme_create_preview (GdkColor *colors)
+{
+    GdkPixbuf *theme_preview;
+    GdkPixmap *drawable;
+    GdkColormap *cmap = gdk_colormap_get_system ();
+    cairo_t *cr;
+    gint width = 44;
+    gint height = 22;
+
+    drawable = gdk_pixmap_new (NULL, width, height, 24);
+    cr = gdk_cairo_create (drawable);
+    cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
+
+    /* Draw three rectangles showcasing the background, foreground and selected background colors */
+    cairo_rectangle (cr, 0, 0, 16, 22);
+    cairo_set_source_rgb (cr, colors[COLOR_BG].red / 65535.0, colors[COLOR_BG].green / 65535.0, colors[COLOR_BG].blue / 65535.0);
+    cairo_fill (cr);
+
+    cairo_rectangle (cr, 15, 0, 30, 22);
+    cairo_set_source_rgb (cr, colors[COLOR_FG].red / 65535.0, colors[COLOR_FG].green / 65535.0, colors[COLOR_FG].blue / 65535.0);
+    cairo_fill (cr);
+
+    cairo_rectangle (cr, 29, 0, 42, 22);
+    cairo_set_source_rgb (cr, colors[COLOR_SELECTED_BG].red / 65535.0, colors[COLOR_SELECTED_BG].green / 65535.0, colors[COLOR_SELECTED_BG].blue / 65535.0);
+    cairo_fill (cr);
+
+    theme_preview = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, width, height);
+    gdk_pixbuf_get_from_drawable (theme_preview, drawable, cmap, 0, 0, 0, 0, width, height);
+
+    g_object_unref (drawable);
+    cairo_destroy (cr);
+
+    return theme_preview;
+}
+
 static void
 cb_theme_tree_selection_changed (GtkTreeSelection *selection,
                                  const gchar      *property)
@@ -166,8 +529,8 @@ cb_toolbar_style_combo_changed (GtkComboBox *combo)
 {
     gint active;
 
-    /* Get active item, prevent number outside the array */
-    active = CLAMP (gtk_combo_box_get_active (combo), 0, (gint) G_N_ELEMENTS (toolbar_styles_array));
+    /* Get active item, prevent number outside the array (stay within zero-index) */
+    active = CLAMP (gtk_combo_box_get_active (combo), 0, (gint) G_N_ELEMENTS (toolbar_styles_array)-1);
 
     /* Save setting */
     xfconf_channel_set_string (xsettings_channel, "/Gtk/ToolbarStyle", toolbar_styles_array[active]);
@@ -193,8 +556,8 @@ cb_hinting_style_combo_changed (GtkComboBox *combo)
 {
     gint active;
 
-    /* Get active item, prevent number outside the array */
-    active = CLAMP (gtk_combo_box_get_active (combo), 0, (gint) G_N_ELEMENTS (xft_hint_styles_array));
+    /* Get active item, prevent number outside the array (stay within zero-index) */
+    active = CLAMP (gtk_combo_box_get_active (combo), 0, (gint) G_N_ELEMENTS (xft_hint_styles_array)-1);
 
     /* Save setting */
     xfconf_channel_set_string (xsettings_channel, "/Xft/HintStyle", xft_hint_styles_array[active]);
@@ -205,8 +568,8 @@ cb_rgba_style_combo_changed (GtkComboBox *combo)
 {
     gint active;
 
-    /* Get active item, prevent number outside the array */
-    active = CLAMP (gtk_combo_box_get_active (combo), 0, (gint) G_N_ELEMENTS (xft_rgba_array));
+    /* Get active item, prevent number outside the array (stay within zero-index) */
+    active = CLAMP (gtk_combo_box_get_active (combo), 0, (gint) G_N_ELEMENTS (xft_rgba_array)-1);
 
     /* Save setting */
     xfconf_channel_set_string (xsettings_channel, "/Xft/RGBA", xft_rgba_array[active]);
@@ -274,10 +637,11 @@ cb_enable_event_sounds_check_button_toggled (GtkToggleButton *toggle, GtkWidget
 }
 #endif
 
-static void
-appearance_settings_load_icon_themes (GtkListStore *list_store,
-                                      GtkTreeView  *tree_view)
+static gboolean
+appearance_settings_load_icon_themes (preview_data *pd)
 {
+    GtkListStore *list_store;
+    GtkTreeView  *tree_view;
     GDir         *dir;
     GtkTreePath  *tree_path;
     GtkTreeIter   iter;
@@ -287,10 +651,26 @@ appearance_settings_load_icon_themes (GtkListStore *list_store,
     gchar        *index_filename;
     const gchar  *theme_name;
     const gchar  *theme_comment;
+    gchar        *name_escaped;
     gchar        *comment_escaped;
+    gchar        *visible_name;
     gchar        *active_theme_name;
-    gint          i;
+    gsize         i;
+    gsize         p;
     GSList       *check_list = NULL;
+    gchar        *cache_filename;
+    gboolean      has_cache;
+    gchar        *cache_tooltip;
+    GtkIconTheme *icon_theme;
+    GdkPixbuf    *preview;
+    GdkPixbuf    *icon;
+    gchar*        preview_icons[4] = { "folder", "go-down", "audio-volume-high", "web-browser" };
+    int           coords[4][2] = { { 4, 4 }, { 24, 4 }, { 4, 24 }, { 24, 24 } };
+
+    g_return_val_if_fail (pd != NULL, FALSE);
+
+    list_store = pd->list_store;
+    tree_view = pd->tree_view;
 
     /* Determine current theme */
     active_theme_name = xfconf_channel_get_string (xsettings_channel, "/Net/IconThemeName", "Rodent");
@@ -332,22 +712,59 @@ appearance_settings_load_icon_themes (GtkListStore *list_store,
                     /* Insert the theme in the check list */
                     check_list = g_slist_prepend (check_list, g_strdup (file));
 
+                    /* Create the icon-theme preview */
+                    preview = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, 44, 44);
+                    gdk_pixbuf_fill (preview, 0x00);
+                    icon_theme = gtk_icon_theme_new ();
+                    gtk_icon_theme_set_custom_theme (icon_theme, file);
+
+                    for (p = 0; p < 4; p++)
+                    {
+                        if (gtk_icon_theme_has_icon (icon_theme, preview_icons[p]))
+                            icon = gtk_icon_theme_load_icon (icon_theme, preview_icons[p], 16, 0, NULL);
+                        else if (gtk_icon_theme_has_icon (icon_theme, "image-missing"))
+                            icon = gtk_icon_theme_load_icon (icon_theme, "image-missing", 16, 0, NULL);
+
+                        if (icon)
+                        {
+                            gdk_pixbuf_copy_area (icon, 0, 0, 16, 16, preview, coords[p][0], coords[p][1]);
+                            g_object_unref (icon);
+                        }
+                    }
+
                     /* Get translated icon theme name and comment */
                     theme_name = xfce_rc_read_entry (index_file, "Name", file);
                     theme_comment = xfce_rc_read_entry (index_file, "Comment", NULL);
 
-                    /* Escape the comment, since tooltips are markup, not text */
+                    /* Escape the theme's name and comment, since they are markup, not text */
+                    name_escaped = g_markup_escape_text (theme_name, -1);
                     comment_escaped = theme_comment ? g_markup_escape_text (theme_comment, -1) : NULL;
+                    visible_name = g_strdup_printf ("<b>%s</b>\n%s", name_escaped, comment_escaped);
+                    g_free (name_escaped);
+                    g_free (comment_escaped);
+
+                    /* Cache filename */
+                    cache_filename = g_build_filename (icon_theme_dirs[i], file, "icon-theme.cache", NULL);
+                    has_cache = g_file_test (cache_filename, G_FILE_TEST_IS_REGULAR);
+                    g_free (cache_filename);
+
+                    /* If the theme has no cache, mention this in the tooltip */
+                    if (!has_cache)
+                        cache_tooltip = g_strdup_printf (_("Warning: this icon theme has no cache file. You can create this by "
+                                                           "running <i>gtk-update-icon-cache %s/%s/</i> in a terminal emulator."),
+                                                         icon_theme_dirs[i], file);
+                    else
+                        cache_tooltip = NULL;
 
                     /* Append icon theme to the list store */
                     gtk_list_store_append (list_store, &iter);
                     gtk_list_store_set (list_store, &iter,
+                                        COLUMN_THEME_PREVIEW, preview,
                                         COLUMN_THEME_NAME, file,
-                                        COLUMN_THEME_DISPLAY_NAME, theme_name,
-                                        COLUMN_THEME_COMMENT, comment_escaped, -1);
-
-                    /* Cleanup */
-                    g_free (comment_escaped);
+                                        COLUMN_THEME_DISPLAY_NAME, visible_name,
+                                        COLUMN_THEME_NO_CACHE, !has_cache,
+                                        COLUMN_THEME_COMMENT, cache_tooltip,
+                                        -1);
 
                     /* Check if this is the active theme, if so, select it */
                     if (G_UNLIKELY (g_utf8_collate (file, active_theme_name) == 0))
@@ -357,6 +774,8 @@ appearance_settings_load_icon_themes (GtkListStore *list_store,
                         gtk_tree_view_scroll_to_cell (tree_view, tree_path, NULL, TRUE, 0.5, 0);
                         gtk_tree_path_free (tree_path);
                     }
+
+                    g_object_unref (preview);
                 }
             }
 
@@ -384,12 +803,15 @@ appearance_settings_load_icon_themes (GtkListStore *list_store,
         g_slist_foreach (check_list, (GFunc) g_free, NULL);
         g_slist_free (check_list);
     }
+
+    return FALSE;
 }
 
-static void
-appearance_settings_load_ui_themes (GtkListStore *list_store,
-                                    GtkTreeView  *tree_view)
+static gboolean
+appearance_settings_load_ui_themes (preview_data *pd)
 {
+    GtkListStore *list_store;
+    GtkTreeView  *tree_view;
     GDir         *dir;
     GtkTreePath  *tree_path;
     GtkTreeIter   iter;
@@ -404,6 +826,14 @@ appearance_settings_load_ui_themes (GtkListStore *list_store,
     gchar        *comment_escaped;
     gint          i;
     GSList       *check_list = NULL;
+    gchar        *color_scheme = NULL;
+    GdkPixbuf    *preview;
+    GdkColor      colors[NUM_SYMBOLIC_COLORS];
+
+    g_return_val_if_fail (pd != NULL, FALSE);
+
+    list_store = pd->list_store;
+    tree_view = pd->tree_view;
 
     /* Determine current theme */
     active_theme_name = xfconf_channel_get_string (xsettings_channel, "/Net/ThemeName", "Default");
@@ -458,9 +888,22 @@ appearance_settings_load_ui_themes (GtkListStore *list_store,
                     comment_escaped = NULL;
                 }
 
+                /* Retrieve the color values from the theme, parse them and create the palette preview pixbuf */
+                color_scheme = gtkrc_get_color_scheme_for_theme (gtkrc_filename);
+                if (color_scheme_parse_colors (color_scheme, colors))
+                    preview = theme_create_preview (colors);
+                /* If the color scheme parsing doesn't return anything useful, show a blank pixbuf */
+                else
+                {
+                    preview = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, 44, 22);
+                    gdk_pixbuf_fill (preview, 0x00);
+                }
+                g_free (color_scheme);
+
                 /* Append ui theme to the list store */
                 gtk_list_store_append (list_store, &iter);
                 gtk_list_store_set (list_store, &iter,
+                                    COLUMN_THEME_PREVIEW, preview,
                                     COLUMN_THEME_NAME, file,
                                     COLUMN_THEME_DISPLAY_NAME, theme_name,
                                     COLUMN_THEME_COMMENT, comment_escaped, -1);
@@ -469,6 +912,7 @@ appearance_settings_load_ui_themes (GtkListStore *list_store,
                 if (G_LIKELY (index_file != NULL))
                     xfce_rc_close (index_file);
                 g_free (comment_escaped);
+                g_object_unref (preview);
 
                 /* Check if this is the active theme, if so, select it */
                 if (G_UNLIKELY (g_utf8_collate (file, active_theme_name) == 0))
@@ -503,6 +947,8 @@ appearance_settings_load_ui_themes (GtkListStore *list_store,
         g_slist_foreach (check_list, (GFunc) g_free, NULL);
         g_slist_free (check_list);
     }
+
+    return FALSE;
 }
 
 static void
@@ -636,8 +1082,15 @@ appearance_settings_dialog_channel_property_changed (XfconfChannel *channel,
 
         if (reload)
         {
+            preview_data *pd;
+
             gtk_list_store_clear (GTK_LIST_STORE (model));
-            appearance_settings_load_ui_themes (GTK_LIST_STORE (model), GTK_TREE_VIEW (object));
+
+            pd = preview_data_new (GTK_LIST_STORE (model), GTK_TREE_VIEW (object));
+            g_idle_add_full (G_PRIORITY_HIGH_IDLE,
+                             (GSourceFunc) appearance_settings_load_ui_themes,
+                             pd,
+                             (GDestroyNotify) preview_data_free);
         }
     }
     else if (strcmp (property_name, "/Net/IconThemeName") == 0)
@@ -670,8 +1123,14 @@ appearance_settings_dialog_channel_property_changed (XfconfChannel *channel,
 
         if (reload)
         {
+            preview_data *pd;
+
             gtk_list_store_clear (GTK_LIST_STORE (model));
-            appearance_settings_load_icon_themes (GTK_LIST_STORE (model), GTK_TREE_VIEW (object));
+            pd = preview_data_new (GTK_LIST_STORE (model), GTK_TREE_VIEW (object));
+            g_idle_add_full (G_PRIORITY_DEFAULT_IDLE,
+                             (GSourceFunc) appearance_settings_load_icon_themes,
+                             pd,
+                             (GDestroyNotify) preview_data_free);
         }
     }
 }
@@ -698,6 +1157,7 @@ cb_theme_uri_dropped (GtkWidget        *widget,
     gboolean       something_installed = FALSE;
     GObject       *object;
     GtkTreeModel  *model;
+    preview_data  *pd;
 
     uris = gtk_selection_data_get_uris (data);
     if (uris == NULL)
@@ -774,41 +1234,75 @@ cb_theme_uri_dropped (GtkWidget        *widget,
 
     if (something_installed)
     {
-        /* reload icon theme treeview */
+        /* reload icon theme treeview in an idle loop */
         object = gtk_builder_get_object (builder, "icon_theme_treeview");
         model = gtk_tree_view_get_model (GTK_TREE_VIEW (object));
         gtk_list_store_clear (GTK_LIST_STORE (model));
-        appearance_settings_load_icon_themes (GTK_LIST_STORE (model), GTK_TREE_VIEW (object));
+        pd = preview_data_new (GTK_LIST_STORE (model), GTK_TREE_VIEW (object));
+        g_idle_add_full (G_PRIORITY_DEFAULT_IDLE,
+                         (GSourceFunc) appearance_settings_load_icon_themes,
+                         pd,
+                         (GDestroyNotify) preview_data_free);
 
         /* reload gtk theme treeview */
         object = gtk_builder_get_object (builder, "gtk_theme_treeview");
         model = gtk_tree_view_get_model (GTK_TREE_VIEW (object));
         gtk_list_store_clear (GTK_LIST_STORE (model));
-        appearance_settings_load_ui_themes (GTK_LIST_STORE (model), GTK_TREE_VIEW (object));
+
+        pd = preview_data_new (GTK_LIST_STORE (model), GTK_TREE_VIEW (object));
+        g_idle_add_full (G_PRIORITY_HIGH_IDLE,
+                         (GSourceFunc) appearance_settings_load_ui_themes,
+                         pd,
+                         (GDestroyNotify) preview_data_free);
     }
 }
 
 static void
 appearance_settings_dialog_configure_widgets (GtkBuilder *builder)
 {
-    GObject          *object, *object2;
-    GtkListStore     *list_store;
-    GtkCellRenderer  *renderer;
-    GdkPixbuf        *pixbuf;
-    GtkTreeSelection *selection;
+    GObject           *object, *object2;
+    GtkListStore      *list_store;
+    GtkCellRenderer   *renderer;
+    GdkPixbuf         *pixbuf;
+    GtkTreeSelection  *selection;
+    GtkTreeViewColumn *column;
+    preview_data      *pd;
 
     /* Icon themes list */
     object = gtk_builder_get_object (builder, "icon_theme_treeview");
 
-    list_store = gtk_list_store_new (N_THEME_COLUMNS, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
+    list_store = gtk_list_store_new (N_THEME_COLUMNS, GDK_TYPE_PIXBUF, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_BOOLEAN);
     gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (list_store), COLUMN_THEME_DISPLAY_NAME, GTK_SORT_ASCENDING);
     gtk_tree_view_set_model (GTK_TREE_VIEW (object), GTK_TREE_MODEL (list_store));
     gtk_tree_view_set_tooltip_column (GTK_TREE_VIEW (object), COLUMN_THEME_COMMENT);
+    gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (object), TRUE);
 
+    /* Single-column layout */
+    column = gtk_tree_view_column_new ();
+
+    /* Icon Previews */
+    renderer = gtk_cell_renderer_pixbuf_new ();
+    gtk_tree_view_column_pack_start (column, renderer, FALSE);
+    gtk_tree_view_column_set_attributes (column, renderer, "pixbuf", COLUMN_THEME_PREVIEW, NULL);
+    gtk_tree_view_append_column (GTK_TREE_VIEW (object), column);
+
+    /* Theme Name and Description */
     renderer = gtk_cell_renderer_text_new ();
-    gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (object), 0, "", renderer, "text", COLUMN_THEME_DISPLAY_NAME, NULL);
+    gtk_tree_view_column_pack_start (column, renderer, TRUE);
+    gtk_tree_view_column_set_attributes (column, renderer, "markup", COLUMN_THEME_DISPLAY_NAME, NULL);
+    g_object_set (G_OBJECT (renderer), "ellipsize", PANGO_ELLIPSIZE_END, NULL);
 
-    appearance_settings_load_icon_themes (list_store, GTK_TREE_VIEW (object));
+    /* Warning Icon */
+    renderer = gtk_cell_renderer_pixbuf_new ();
+    gtk_tree_view_column_pack_start (column, renderer, FALSE);
+    gtk_tree_view_column_set_attributes (column, renderer, "visible", COLUMN_THEME_NO_CACHE, NULL);
+    g_object_set (G_OBJECT (renderer), "icon-name", GTK_STOCK_DIALOG_WARNING, NULL);
+
+    pd = preview_data_new (GTK_LIST_STORE (list_store), GTK_TREE_VIEW (object));
+    g_idle_add_full (G_PRIORITY_DEFAULT_IDLE,
+                     (GSourceFunc) appearance_settings_load_icon_themes,
+                     pd,
+                     (GDestroyNotify) preview_data_free);
 
     g_object_unref (G_OBJECT (list_store));
 
@@ -824,15 +1318,31 @@ appearance_settings_dialog_configure_widgets (GtkBuilder *builder)
     /* Gtk (UI) themes */
     object = gtk_builder_get_object (builder, "gtk_theme_treeview");
 
-    list_store = gtk_list_store_new (N_THEME_COLUMNS, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
+    list_store = gtk_list_store_new (N_THEME_COLUMNS, GDK_TYPE_PIXBUF, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_BOOLEAN);
     gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (list_store), COLUMN_THEME_DISPLAY_NAME, GTK_SORT_ASCENDING);
     gtk_tree_view_set_model (GTK_TREE_VIEW (object), GTK_TREE_MODEL (list_store));
     gtk_tree_view_set_tooltip_column (GTK_TREE_VIEW (object), COLUMN_THEME_COMMENT);
 
-    renderer = gtk_cell_renderer_text_new();
-    gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (object), 0, "", renderer, "text", COLUMN_THEME_DISPLAY_NAME, NULL);
+    /* Single-column layout */
+    column = gtk_tree_view_column_new ();
 
-    appearance_settings_load_ui_themes (list_store, GTK_TREE_VIEW (object));
+    /* Icon Previews */
+    renderer = gtk_cell_renderer_pixbuf_new ();
+    gtk_tree_view_column_pack_start (column, renderer, FALSE);
+    gtk_tree_view_column_set_attributes (column, renderer, "pixbuf", COLUMN_THEME_PREVIEW, NULL);
+    gtk_tree_view_append_column (GTK_TREE_VIEW (object), column);
+
+    /* Theme Name and Description */
+    renderer = gtk_cell_renderer_text_new ();
+    gtk_tree_view_column_pack_start (column, renderer, TRUE);
+    gtk_tree_view_column_set_attributes (column, renderer, "text", COLUMN_THEME_DISPLAY_NAME, NULL);
+    g_object_set (G_OBJECT (renderer), "ellipsize", PANGO_ELLIPSIZE_END, NULL);
+
+    pd = preview_data_new (list_store, GTK_TREE_VIEW (object));
+    g_idle_add_full (G_PRIORITY_HIGH_IDLE,
+                     (GSourceFunc) appearance_settings_load_ui_themes,
+                     pd,
+                     (GDestroyNotify) preview_data_free);
 
     g_object_unref (G_OBJECT (list_store));
 
@@ -940,7 +1450,8 @@ appearance_settings_dialog_response (GtkWidget *dialog,
                                      gint       response_id)
 {
     if (response_id == GTK_RESPONSE_HELP)
-        xfce_dialog_show_help (GTK_WINDOW (dialog), "xfce4-settings", "appearance", NULL);
+        xfce_dialog_show_help_with_version (GTK_WINDOW (dialog), "xfce4-settings", "appearance",
+                                            NULL, XFCE4_SETTINGS_VERSION_SHORT);
     else
         gtk_main_quit ();
 }
